-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      CipherCore for NORX
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use ieee.std_logic_unsigned.ALL;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 768;   --! Data
        G_KEY_SIZE      : integer := 32;    --! Key
        G_TAG_SIZE      : integer := 256;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 6;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX;
        --! NORX parameters
        G_NORX_L        : integer := 4
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
    --! =======================================================================
    --! Datapath
    --! =======================================================================
    --! Constants
    --!     NORX parameters
    constant NORX_W     : integer := (G_DBLK_SIZE/384)*32;
    constant NORX_WB    : integer := NORX_W/8;
    constant NORX_L     : integer := G_NORX_L;
    constant NORX_K     : integer := NORX_W * 4;
    constant NORX_T     : integer := NORX_W * 4;
    constant NORX_N     : integer := NORX_W * 4;
    constant NORX_P     : integer := 1;
    --!     NORX S parameters
    constant NORX_B     : integer := NORX_W * 16;
    constant NORX_C     : integer := NORX_W *  4;
    constant NORX_R     : integer := NORX_B - NORX_C;

    constant ZEROS      : std_logic_vector(NORX_B downto 0) := (others => '0');

    type t_rot_offset is array (1 to 2, 0 to 3) of integer;
    constant ROT_OFFSET : t_rot_offset := ((8, 11, 16, 31),     --! W=32
                                           (8, 19, 40, 63));    --! W=64

    --! Initial values
    constant INIT_W64_S08 : std_logic_vector(63 downto 0) := x"B15E641748DE5E6B";
    constant INIT_W64_S09 : std_logic_vector(63 downto 0) := x"AA95E955E10F8410";
    constant INIT_W64_S10 : std_logic_vector(63 downto 0) := x"28D1034441A9DD40";
    constant INIT_W64_S11 : std_logic_vector(63 downto 0) := x"7F31BBF964E93BF5";
    constant INIT_W64_S12 : std_logic_vector(63 downto 0) := x"B5E9E22493DFFB96" xor std_logic_vector(to_unsigned(NORX_W, 64));
    constant INIT_W64_S13 : std_logic_vector(63 downto 0) := x"B980C852479FAFBD" xor std_logic_vector(to_unsigned(NORX_L, 64));
    constant INIT_W64_S14 : std_logic_vector(63 downto 0) := x"DA24516BF55EAFD4" xor std_logic_vector(to_unsigned(NORX_P, 64));
    constant INIT_W64_S15 : std_logic_vector(63 downto 0) := x"86026AE8536F1501" xor std_logic_vector(to_unsigned(NORX_T, 64));    
    constant INIT_W32_S08 : std_logic_vector(31 downto 0) := x"A3D8D930";
    constant INIT_W32_S09 : std_logic_vector(31 downto 0) := x"3FA8B72C";
    constant INIT_W32_S10 : std_logic_vector(31 downto 0) := x"ED84EB49";
    constant INIT_W32_S11 : std_logic_vector(31 downto 0) := x"EDCA4787";
    constant INIT_W32_S12 : std_logic_vector(31 downto 0) := x"335463EB" xor std_logic_vector(to_unsigned(NORX_W, 32));
    constant INIT_W32_S13 : std_logic_vector(31 downto 0) := x"F994220B" xor std_logic_vector(to_unsigned(NORX_L, 32));
    constant INIT_W32_S14 : std_logic_vector(31 downto 0) := x"BE0BF5C9" xor std_logic_vector(to_unsigned(NORX_P, 32));
    constant INIT_W32_S15 : std_logic_vector(31 downto 0) := x"D7C49104" xor std_logic_vector(to_unsigned(NORX_T, 32));

    function switch_endian (
        signal      x       : std_logic_vector;
        constant    size    : integer;
        constant    w       : integer
    ) return std_logic_vector is
        constant xlow   : integer := x'low;
		constant xseg   : integer := size/w;
		constant wseg   : integer := w/8;
		variable retval : std_logic_vector(size-1 downto 0);
	begin
		for i in xseg downto 1 loop
			for j in 0 to wseg-1 loop
				retval(i*w-j*8+xlow-1 downto i*w-j*8+xlow-8)
                    := x(i*w-(wseg-1-j)*8+xlow-1
                         downto i*w-(wseg-1-j)*8+xlow-8);
			end loop;
		end loop;
		return(retval);
	end switch_endian;

    function FF (
        x : unsigned(NORX_W-1 downto 0);
        y : unsigned(NORX_W-1 downto 0)
    ) return unsigned is
        variable ret : unsigned(NORX_W-1 downto 0);
    begin
        ret := (x xor y) xor shift_left((x and y), 1);
        return ret;
    end FF;

    procedure GG (
        signal a,b,c,d : in  std_logic_vector(NORX_W        -1 downto 0);
        signal w,x,y,z : out std_logic_vector(NORX_W        -1 downto 0))
    is
        variable tmp_a, tmp_b, tmp_c, tmp_d : unsigned(NORX_W   -1 downto 0);
        variable tmp_a2, tmp_b2             : unsigned(NORX_W   -1 downto 0);
        variable tmp_c2, tmp_d2             : unsigned(NORX_W   -1 downto 0);
        variable rot_d, rot_b               : unsigned(NORX_W   -1 downto 0);
        variable rot_d2, rot_b2             : unsigned(NORX_W   -1 downto 0);
    begin
        --! a = H(a,b); d ^= a; d >>>= r0;
        tmp_a := FF(unsigned(a), unsigned(b));
        tmp_d := unsigned(d) xor tmp_a;
        rot_d := rotate_right(tmp_d, ROT_OFFSET(NORX_W/32, 0));

        --! c = H(c,d); b ^= c; d >>>= r1;
        tmp_c := FF(unsigned(c), rot_d);
        tmp_b := unsigned(b) xor tmp_c;
        rot_b := rotate_right(tmp_b, ROT_OFFSET(NORX_W/32, 1));

        --! a = H(a,b); d ^= a; d >>>= r2;
        tmp_a2 := FF(tmp_a, rot_b);
        tmp_d2 := rot_d xor tmp_a2;
        rot_d2 := rotate_right(tmp_d2, ROT_OFFSET(NORX_W/32, 2));

        --! c = H(c,d); b ^= c; d >>>= r3;
        tmp_c2 := FF(tmp_c, rot_d2);
        tmp_b2 := rot_b xor tmp_c2;
        rot_b2 := rotate_right(tmp_b2, ROT_OFFSET(NORX_W/32, 3));

        w <= std_logic_vector(tmp_a2);
        x <= std_logic_vector(rot_b2);
        y <= std_logic_vector(tmp_c2);
        z <= std_logic_vector(rot_d2);
    end GG;

    signal bdi_valid_bytes_le   : std_logic_vector(NORX_R/8     -1 downto 0);
    signal key_r                : std_logic_vector(NORX_K       -1 downto 0);
    signal keyinj               : std_logic_vector(NORX_K       -1 downto 0);
    signal input                : std_logic_vector(NORX_R       -1 downto 0);
    signal output               : std_logic_vector(NORX_R       -1 downto 0);
    signal to_state             : std_logic_vector(NORX_B       -1 downto 0);
    signal init_state           : std_logic_vector(NORX_B       -1 downto 0);
    signal new_enc_state        : std_logic_vector(NORX_B       -1 downto 0);
    signal new_dec_state        : std_logic_vector(NORX_B       -1 downto 0);
    signal state                : std_logic_vector(NORX_B       -1 downto 0);
    signal from_round_pre       : std_logic_vector(NORX_B       -1 downto 0);
    signal from_round           : std_logic_vector(NORX_B       -1 downto 0);
    signal from_round_or_final  : std_logic_vector(NORX_B       -1 downto 0);
    signal tag_le               : std_logic_vector(NORX_T       -1 downto 0);
    signal tag                  : std_logic_vector(NORX_T       -1 downto 0);
    

    type t_array is array (0 to 15) of std_logic_vector(NORX_W-1 downto 0);
    signal ss           : t_array;
    signal col          : t_array;
    signal diag         : t_array;

    --! =======================================================================
    --! Control
    --! =======================================================================
    type t_state is (
        S_INIT, S_LDKEY, S_LD_NPUB,
        S_LDBDI, S_PROCESS, S_FINALIZE, S_TAG);
    signal cs, ns       : t_state;

    --!     External
    signal key_rdy      : std_logic;
    signal bdi_rdy      : std_logic;
    signal bdo_vld      : std_logic;
    --!     Datapath
    --!         Controls
    signal sel_input    : std_logic_vector(2            -1 downto 0);
    signal sel_domain   : std_logic_vector(2            -1 downto 0);
    signal sel_keyinj   : std_logic;
    signal sel_final    : std_logic;
    signal sel_tag      : std_logic;
    signal sel_pad      : std_logic;
    signal en_state     : std_logic;

    --!     Internal
    signal ld_rnd       : std_logic;
    signal en_rnd       : std_logic;
    signal rnd          : std_logic_vector(1+NORX_L/2   -1 downto 0);
    --!         Status
    signal is_eoi       : std_logic;
    signal is_init      : std_logic;
    signal is_decrypt   : std_logic;
begin
    --! =======================================================================
    --! Datapath
    --! =======================================================================
    input <= switch_endian(bdi, NORX_R, NORX_W);
    genNorxW64:
    if (NORX_W = 64) generate
        init_state <=
            input(NORX_R-1 downto NORX_R-NORX_N)
            & switch_endian(key_r, NORX_K, NORX_W)
            & INIT_W64_S08 & INIT_W64_S09 & INIT_W64_S10 & INIT_W64_S11
            & INIT_W64_S12 & INIT_W64_S13 & INIT_W64_S14 & INIT_W64_S15;
    end generate;
    genNorxW32:
    if (NORX_W = 32) generate
        init_state <=
            input(NORX_R-1 downto NORX_R-NORX_N)
            & switch_endian(key_r, NORX_K, NORX_W)
            & INIT_W32_S08 & INIT_W32_S09 & INIT_W32_S10 & INIT_W32_S11
            & INIT_W32_S12 & INIT_W32_S13 & INIT_W32_S14 & INIT_W32_S15;
    end generate;

    output <= from_round(NORX_B-1 downto NORX_C) xor input;
    tag_le <= from_round(NORX_T-1 downto 0);
    tag    <= switch_endian(tag_le, NORX_T, NORX_W);
    bdo    <= tag & ZEROS(NORX_R-NORX_T-1 downto 0)
              when sel_tag = '1'
              else switch_endian(output, NORX_R, NORX_W);
    msg_auth_valid <= '1'
                      when tag = bdi(NORX_R-1 downto NORX_R-NORX_T)
                      else '0';


    new_enc_state(NORX_B-1 downto NORX_C) <= output;
    new_enc_state(NORX_C-1 downto 0) <= from_round(NORX_C-1 downto 0);
    
    genNorxEncDec:
    for i in NORX_R/8-1 downto 0 generate
        bdi_valid_bytes_le((NORX_WB-1-(i mod NORX_WB))+(i/NORX_WB)*NORX_WB) 
            <= bdi_valid_bytes(i);
        
        new_dec_state(NORX_C+(i+1)*8-1 downto NORX_C+i*8)
            <= input((i+1)*8-1 downto i*8)
               when bdi_valid_bytes_le(i) = '1'
               else output((i+1)*8-1 downto i*8);
    end generate;
    new_dec_state(NORX_C-1 downto 0)  <= from_round(NORX_C-1 downto 0);

    with sel_input select
    to_state <= from_round    when "00",
                init_state    when "01",
                new_enc_state when "10",
                new_dec_state when others;

    --! Round
    GG(ss( 0),ss( 4),ss( 8),ss(12),col( 0),col( 4),col( 8),col(12));
    GG(ss( 1),ss( 5),ss( 9),ss(13),col( 1),col( 5),col( 9),col(13));
    GG(ss( 2),ss( 6),ss(10),ss(14),col( 2),col( 6),col(10),col(14));
    GG(ss( 3),ss( 7),ss(11),ss(15),col( 3),col( 7),col(11),col(15));
    GG(col( 0),col( 5),col(10),col(15),diag( 0),diag( 5),diag(10),diag(15));
    GG(col( 1),col( 6),col(11),col(12),diag( 1),diag( 6),diag(11),diag(12));
    GG(col( 2),col( 7),col( 8),col(13),diag( 2),diag( 7),diag( 8),diag(13));
    GG(col( 3),col( 4),col( 9),col(14),diag( 3),diag( 4),diag( 9),diag(14));

    genArray:
    for i in 15 downto 0 generate
        ss(15-i) <= state((i+1)*NORX_W-1 downto i*NORX_W);
        from_round_pre((i+1)*NORX_W-1 downto i*NORX_W) <= diag(15-i);
    end generate;
    
    from_round(NORX_B-1 downto NORX_K) <= from_round_pre(NORX_B-1 downto NORX_K);
    keyinj                             <= switch_endian(key_r, NORX_K, NORX_W) 
                                           when sel_keyinj = '1' else (others => '0');
    from_round(NORX_K-1 downto 0)      <= from_round_pre(NORX_K-1 downto 0) xor keyinj;

    --! Registers
    regs:
    process(clk)
    begin
        if rising_edge(clk) then
            if key_rdy = '1' then
                key_r <= key_r(NORX_K-32-1 downto 0) & key;
            end if;

            if en_state = '1' then
                state(NORX_B -1       downto NORX_C+NORX_W  ) <= to_state(NORX_B -1 downto NORX_C+NORX_W);
                state(                       NORX_C+NORX_W-1) <= to_state(NORX_C+NORX_W-1) xor sel_pad;
                state(NORX_C+NORX_W-2 downto               4) <= to_state(NORX_C+NORX_W-2 downto 4);
                state(                                     3) <= to_state(3) xor sel_final;
                state(                                     2) <= to_state(2);
                state(              1 downto               0) <= to_state(1 downto 0) xor sel_domain;
            end if;
        end if;
    end process;

    --! =======================================================================
    --! Controller
    --! =======================================================================
    ctrlregs:
    process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                cs <= S_INIT;
            else
                cs <= ns;
            end if;

            if (ld_rnd = '1') then
                rnd <= (others => '0');
            elsif (en_rnd = '1') then
                rnd <= std_logic_vector(unsigned(rnd) + 1);
            end if;

            if (bdi_rdy = '1') then
                is_eoi <= bdi_eoi;
            end if;

            if (cs = S_LD_NPUB) then
                sel_tag <= '0';
            elsif (cs = S_FINALIZE) then
                sel_tag <= '1';
            end if;
            
            if (cs = S_LD_NPUB) then
                is_init <= '0';
            elsif (cs = S_LDBDI and bdi_valid = '1') then
                is_init <= '1';
            end if;

            if (cs = S_INIT) then
                is_decrypt <= decrypt;
            end if;
        end if;
    end process;

    key_ready <= key_rdy;
    bdi_ready <= bdi_rdy;
    bdo_valid <= bdo_vld;

    comb:
    process(
        cs, bdi_valid, key_update, key_valid, rnd, bdi_eoi, bdi_eot, 
        is_init, bdi_type, bdi_valid_bytes, bdo_ready, is_eoi)
    begin
        ns          <= cs;
        --! External
        key_rdy     <= '0';
        bdi_rdy     <= '0';
        bdo_vld     <= '0';
        msg_auth_done <= '0';
        --! Datapath
        sel_input   <= "00";
        sel_keyinj  <= '0';
        sel_domain  <= "00";
        sel_final   <= '0';
        sel_pad     <= '0';        
        en_state    <= '0';
        --! Internal
        ld_rnd      <= '0';
        en_rnd      <= '0';

        case cs is
        when S_INIT =>
            ld_rnd <= '1';
            if (bdi_valid = '1') then
                if (key_update = '1') then
                    ns <= S_LDKEY;
                else
                    ns <= S_LD_NPUB;
                end if;
            end if;

        when S_LDKEY =>
            if (key_valid = '1') then
                en_rnd  <= '1';
                key_rdy <= '1';
                if (unsigned(rnd) = NORX_K/32-1) then --! 8 words
                    ns <= S_LD_NPUB;
                    ld_rnd <= '1';
                end if;
            end if;

        when S_LD_NPUB =>
            bdi_rdy   <= '1';
            sel_input <= "01";
            if (bdi_valid = '1') then
                en_state <= '1';
                ns       <= S_PROCESS;
            end if;

        when S_LDBDI =>
            ld_rnd  <= '1';
            if (is_eoi = '1') then
                sel_final <= '1';
                en_state  <= '1';                
                ns        <= S_FINALIZE;
                sel_keyinj <= '1';
            elsif (is_init = '0') then
                sel_keyinj <= '1';
                if (bdi_valid = '1') then
                    if (bdi_type(2 downto 1) = BDI_TYPE_DAT) then
                        sel_domain <= "10";
                    else
                        sel_domain <= "01";
                    end if;
                    en_state  <= '1';
                    ns  <= S_PROCESS;
                end if;
            elsif (bdi_valid = '1') then
                sel_input(1) <= '1';
                if (bdi_eot = '1') then
                    sel_pad <= '1';
                end if;
                if (bdi_type(2 downto 1) = BDI_TYPE_DAT
                    and (bdo_ready = '1' or bdi_valid_bytes(NORX_R/8-1) = '0'))
                then
                    if (bdi_eoi = '1') then
                        sel_final  <= '1';                        
                        ns         <= S_FINALIZE;
                    else
                        sel_domain <= "10";
                        ns         <= S_PROCESS;
                    end if;                    
                    sel_input(0) <= is_decrypt;
                    bdi_rdy  <= '1';                    
                    bdo_vld  <= bdi_valid_bytes(NORX_R/8-1);    --! Not empty
                    en_state <= '1';
                    
                elsif (bdi_type(2 downto 1) = BDI_TYPE_ASS) then
                    if (bdi_eoi = '1') then
                        sel_final    <= '1';                        
                        ns           <= S_FINALIZE;
                    else
                        if (bdi_eot = '1') then
                            sel_domain <= "10";
                        else
                            sel_domain <= "01";
                        end if;
                        ns           <= S_PROCESS;
                    end if;
                    bdi_rdy  <= '1';
                    en_state <= '1';                    
                end if;
            end if;

        when S_PROCESS =>
            en_rnd   <= '1';
            en_state <= '1';
            if (unsigned(rnd) = NORX_L-2) then
                ns        <= S_LDBDI;
            end if;

        when S_FINALIZE =>
            en_rnd   <= '1';
            en_state <= '1';
            if (unsigned(rnd) = (2*NORX_L)-2) then
                ns      <= S_TAG;
            end if;
            if (unsigned(rnd) = NORX_L-1) then
                sel_keyinj   <= '1';
            end if;

        when S_TAG =>
            sel_keyinj <= '1';
            if (is_decrypt = '1') then
                bdi_rdy <= '1';
                if (bdi_valid = '1') then
                    msg_auth_done <= '1';
                    ns <= S_INIT;
                end if;
            elsif (bdo_ready = '1') then
                bdo_vld <= '1';
                ns <= S_INIT;
            end if;

        end case;
    end process;

end structure;